"""Defines the unit tests for the :mod:`colour.continuous.multi_signals` module."""

import numpy as np
import unittest
import re
import textwrap

from colour.algebra import (
    CubicSplineInterpolator,
    Extrapolator,
    KernelInterpolator,
)
from colour.constants import DEFAULT_FLOAT_DTYPE
from colour.continuous import MultiSignals, Signal
from colour.utilities import (
    ColourRuntimeWarning,
    attest,
    is_pandas_installed,
    tsplit,
    tstack,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestMultiSignals",
]


class TestMultiSignals(unittest.TestCase):
    """
    Define :class:`colour.continuous.multi_signals.MultiSignals` class unit
    tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._range_1 = np.linspace(10, 100, 10)
        self._range_2 = tstack([self._range_1] * 3) + np.array([0, 10, 20])
        self._domain_1 = np.arange(0, 10, 1)
        self._domain_2 = np.arange(100, 1100, 100)

        self._multi_signals = MultiSignals(self._range_2)

    def test_required_attributes(self):
        """Test the presence of required attributes."""

        required_attributes = (
            "dtype",
            "domain",
            "range",
            "interpolator",
            "interpolator_kwargs",
            "extrapolator",
            "extrapolator_kwargs",
            "function",
            "signals",
            "labels",
            "signal_type",
        )

        for attribute in required_attributes:
            self.assertIn(attribute, dir(MultiSignals))

    def test_required_methods(self):
        """Test the presence of required methods."""

        required_methods = (
            "__init__",
            "__str__",
            "__repr__",
            "__hash__",
            "__getitem__",
            "__setitem__",
            "__contains__",
            "__eq__",
            "__ne__",
            "arithmetical_operation",
            "multi_signals_unpack_data",
            "fill_nan",
            "domain_distance",
            "to_dataframe",
        )

        for method in required_methods:
            self.assertIn(method, dir(MultiSignals))

    def test_dtype(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.dtype`
        property.
        """

        self.assertEqual(self._multi_signals.dtype, DEFAULT_FLOAT_DTYPE)

        multi_signals = self._multi_signals.copy()
        multi_signals.dtype = np.float32
        self.assertEqual(multi_signals.dtype, np.float32)

    def test_domain(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.domain`
        property.
        """

        multi_signals = self._multi_signals.copy()

        np.testing.assert_almost_equal(
            multi_signals[np.array([0, 1, 2])],
            np.array(
                [[10.0, 20.0, 30.0], [20.0, 30.0, 40.0], [30.0, 40.0, 50.0]]
            ),
            decimal=7,
        )

        multi_signals.domain = self._domain_1 * 10

        np.testing.assert_array_equal(
            multi_signals.domain, self._domain_1 * 10
        )

        np.testing.assert_almost_equal(
            multi_signals[np.array([0, 1, 2]) * 10],
            np.array(
                [[10.0, 20.0, 30.0], [20.0, 30.0, 40.0], [30.0, 40.0, 50.0]]
            ),
            decimal=7,
        )

        domain = np.linspace(0, 1, 10)
        domain[0] = -np.inf

        def assert_warns():
            """Help to test the runtime warning."""

            multi_signals.domain = domain

        self.assertWarns(ColourRuntimeWarning, assert_warns)

    def test_range(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.range`
        property.
        """

        multi_signals = self._multi_signals.copy()

        np.testing.assert_almost_equal(
            multi_signals[np.array([0, 1, 2])],
            np.array(
                [[10.0, 20.0, 30.0], [20.0, 30.0, 40.0], [30.0, 40.0, 50.0]]
            ),
            decimal=7,
        )

        multi_signals.range = self._range_1 * 10

        np.testing.assert_array_equal(
            multi_signals.range, tstack([self._range_1] * 3) * 10
        )

        np.testing.assert_almost_equal(
            multi_signals[np.array([0, 1, 2])],
            np.array(
                [[10.0, 10.0, 10.0], [20.0, 20.0, 20.0], [30.0, 30.0, 30.0]]
            )
            * 10,
            decimal=7,
        )

        multi_signals.range = self._range_2 * 10

        np.testing.assert_array_equal(multi_signals.range, self._range_2 * 10)

        np.testing.assert_almost_equal(
            multi_signals[np.array([0, 1, 2])],
            np.array(
                [[10.0, 20.0, 30.0], [20.0, 30.0, 40.0], [30.0, 40.0, 50.0]]
            )
            * 10,
            decimal=7,
        )

    def test_interpolator(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.interpolator`
        property.
        """

        multi_signals = self._multi_signals.copy()

        np.testing.assert_almost_equal(
            multi_signals[np.linspace(0, 5, 5)],
            np.array(
                [
                    [10.00000000, 20.00000000, 30.00000000],
                    [22.83489024, 32.80460562, 42.77432100],
                    [34.80044921, 44.74343470, 54.68642018],
                    [47.55353925, 57.52325463, 67.49297001],
                    [60.00000000, 70.00000000, 80.00000000],
                ]
            ),
            decimal=7,
        )

        multi_signals.interpolator = CubicSplineInterpolator

        np.testing.assert_almost_equal(
            multi_signals[np.linspace(0, 5, 5)],
            np.array(
                [
                    [10.00000000, 20.00000000, 30.00000000],
                    [22.50000000, 32.50000000, 42.50000000],
                    [35.00000000, 45.00000000, 55.00000000],
                    [47.50000000, 57.50000000, 67.50000000],
                    [60.00000000, 70.00000000, 80.00000000],
                ]
            ),
            decimal=7,
        )

    def test_interpolator_kwargs(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.\
interpolator_kwargs` property.
        """

        multi_signals = self._multi_signals.copy()

        np.testing.assert_almost_equal(
            multi_signals[np.linspace(0, 5, 5)],
            np.array(
                [
                    [10.00000000, 20.00000000, 30.00000000],
                    [22.83489024, 32.80460562, 42.77432100],
                    [34.80044921, 44.74343470, 54.68642018],
                    [47.55353925, 57.52325463, 67.49297001],
                    [60.00000000, 70.00000000, 80.00000000],
                ]
            ),
            decimal=7,
        )

        multi_signals.interpolator_kwargs = {
            "window": 1,
            "kernel_kwargs": {"a": 1},
        }

        np.testing.assert_almost_equal(
            multi_signals[np.linspace(0, 5, 5)],
            np.array(
                [
                    [10.00000000, 20.00000000, 30.00000000],
                    [18.91328761, 27.91961505, 36.92594248],
                    [28.36993142, 36.47562611, 44.58132080],
                    [44.13100443, 53.13733187, 62.14365930],
                    [60.00000000, 70.00000000, 80.00000000],
                ]
            ),
            decimal=7,
        )

    def test_extrapolator(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.extrapolator`
        property.
        """

        self.assertIsInstance(self._multi_signals.extrapolator(), Extrapolator)

    def test_extrapolator_kwargs(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.\
extrapolator_kwargs` property.
        """

        multi_signals = self._multi_signals.copy()

        attest(np.all(np.isnan(multi_signals[np.array([-1000, 1000])])))

        multi_signals.extrapolator_kwargs = {
            "method": "Linear",
        }

        np.testing.assert_almost_equal(
            multi_signals[np.array([-1000, 1000])],
            np.array(
                [[-9990.0, -9980.0, -9970.0], [10010.0, 10020.0, 10030.0]]
            ),
            decimal=7,
        )

    def test_function(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.function`
        property.
        """

        attest(hasattr(self._multi_signals.function, "__call__"))

    def test_raise_exception_function(self):
        """
        Test :func:`colour.continuous.signal.multi_signals.MultiSignals.\
function` property raised exception.
        """

        self.assertRaises((ValueError, TypeError), MultiSignals().function, 0)

    def test_signals(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.signals`
        property.
        """

        multi_signals = self._multi_signals.copy()

        multi_signals.signals = self._range_1
        np.testing.assert_array_equal(multi_signals.domain, self._domain_1)
        np.testing.assert_array_equal(
            multi_signals.range, self._range_1[:, np.newaxis]
        )

    def test_labels(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.labels`
        property.
        """

        self.assertListEqual(self._multi_signals.labels, ["0", "1", "2"])

        multi_signals = self._multi_signals.copy()

        multi_signals.labels = ["a", "b", "c"]

        self.assertListEqual(multi_signals.labels, ["a", "b", "c"])

    def test_signal_type(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.signal_type`
        property.
        """

        multi_signals = MultiSignals(signal_type=Signal)

        self.assertEqual(multi_signals.signal_type, Signal)

    def test__init__(self):
        """
        Test :meth:`colour.continuous.multi_signals.MultiSignals.__init__`
        method.
        """

        multi_signals = MultiSignals(self._range_1)
        np.testing.assert_array_equal(multi_signals.domain, self._domain_1)
        np.testing.assert_array_equal(
            multi_signals.range, self._range_1[:, np.newaxis]
        )

        multi_signals = MultiSignals(self._range_1, self._domain_2)
        np.testing.assert_array_equal(multi_signals.domain, self._domain_2)
        np.testing.assert_array_equal(
            multi_signals.range, self._range_1[:, np.newaxis]
        )

        multi_signals = MultiSignals(self._range_2, self._domain_2)
        np.testing.assert_array_equal(multi_signals.domain, self._domain_2)
        np.testing.assert_array_equal(multi_signals.range, self._range_2)

        multi_signals = MultiSignals(dict(zip(self._domain_2, self._range_2)))
        np.testing.assert_array_equal(multi_signals.domain, self._domain_2)
        np.testing.assert_array_equal(multi_signals.range, self._range_2)

        multi_signals = MultiSignals(multi_signals)
        np.testing.assert_array_equal(multi_signals.domain, self._domain_2)
        np.testing.assert_array_equal(multi_signals.range, self._range_2)

        class NotSignal(Signal):
            """Not :class:`Signal` class."""

            pass

        multi_signals = MultiSignals(self._range_1, signal_type=NotSignal)
        self.assertIsInstance(multi_signals.signals["0"], NotSignal)
        np.testing.assert_array_equal(multi_signals.domain, self._domain_1)
        np.testing.assert_array_equal(
            multi_signals.range, self._range_1[:, np.newaxis]
        )

        if is_pandas_installed():
            from pandas import DataFrame, Series

            multi_signals = MultiSignals(
                Series(dict(zip(self._domain_2, self._range_1)))
            )
            np.testing.assert_array_equal(multi_signals.domain, self._domain_2)
            np.testing.assert_array_equal(
                multi_signals.range, self._range_1[:, np.newaxis]
            )

            data = dict(zip(["a", "b", "c"], tsplit(self._range_2)))
            multi_signals = MultiSignals(DataFrame(data, self._domain_2))
            np.testing.assert_array_equal(multi_signals.domain, self._domain_2)
            np.testing.assert_array_equal(multi_signals.range, self._range_2)

    def test__hash__(self):
        """
        Test :meth:`colour.continuous.multi_signals.MultiSignals.__hash__`
        method.
        """

        self.assertIsInstance(hash(self._multi_signals), int)

    def test__str__(self):
        """
        Test :meth:`colour.continuous.multi_signals.MultiSignals.__str__`
        method.
        """

        self.assertEqual(
            str(self._multi_signals),
            textwrap.dedent(
                """
                [[   0.   10.   20.   30.]
                 [   1.   20.   30.   40.]
                 [   2.   30.   40.   50.]
                 [   3.   40.   50.   60.]
                 [   4.   50.   60.   70.]
                 [   5.   60.   70.   80.]
                 [   6.   70.   80.   90.]
                 [   7.   80.   90.  100.]
                 [   8.   90.  100.  110.]
                 [   9.  100.  110.  120.]]"""
            )[1:],
        )

        self.assertIsInstance(str(MultiSignals()), str)

    def test__repr__(self):
        """
        Test :meth:`colour.continuous.multi_signals.MultiSignals.__repr__`
        method.
        """

        self.assertEqual(
            re.sub(
                r"extrapolator_kwargs={.*}",
                "extrapolator_kwargs={...}",
                repr(self._multi_signals),
            ),
            textwrap.dedent(
                """
                MultiSignals([[   0.,   10.,   20.,   30.],
                              [   1.,   20.,   30.,   40.],
                              [   2.,   30.,   40.,   50.],
                              [   3.,   40.,   50.,   60.],
                              [   4.,   50.,   60.,   70.],
                              [   5.,   60.,   70.,   80.],
                              [   6.,   70.,   80.,   90.],
                              [   7.,   80.,   90.,  100.],
                              [   8.,   90.,  100.,  110.],
                              [   9.,  100.,  110.,  120.]],
                             labels=['0', '1', '2'],
                             interpolator=KernelInterpolator,
                             interpolator_kwargs={},
                             extrapolator=Extrapolator,
                             extrapolator_kwargs={...})"""
            )[1:],
        )

        self.assertIsInstance(repr(MultiSignals()), str)

    def test__getitem__(self):
        """
        Test :meth:`colour.continuous.multi_signals.MultiSignals.__getitem__`
        method.
        """

        np.testing.assert_almost_equal(
            self._multi_signals[0], np.array([10.0, 20.0, 30.0]), decimal=7
        )

        np.testing.assert_almost_equal(
            self._multi_signals[np.array([0, 1, 2])],
            np.array(
                [
                    [10.0, 20.0, 30.0],
                    [20.0, 30.0, 40.0],
                    [30.0, 40.0, 50.0],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            self._multi_signals[np.linspace(0, 5, 5)],
            np.array(
                [
                    [10.00000000, 20.00000000, 30.00000000],
                    [22.83489024, 32.80460562, 42.77432100],
                    [34.80044921, 44.74343470, 54.68642018],
                    [47.55353925, 57.52325463, 67.49297001],
                    [60.00000000, 70.00000000, 80.00000000],
                ]
            ),
            decimal=7,
        )

        attest(np.all(np.isnan(self._multi_signals[np.array([-1000, 1000])])))

        np.testing.assert_almost_equal(
            self._multi_signals[:], self._multi_signals.range, decimal=7
        )

        np.testing.assert_almost_equal(
            self._multi_signals[:, :], self._multi_signals.range, decimal=7
        )

        np.testing.assert_almost_equal(
            self._multi_signals[0:3],
            np.array(
                [
                    [10.0, 20.0, 30.0],
                    [20.0, 30.0, 40.0],
                    [30.0, 40.0, 50.0],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            self._multi_signals[:, 0:2],
            np.array(
                [
                    [10.0, 20.0],
                    [20.0, 30.0],
                    [30.0, 40.0],
                    [40.0, 50.0],
                    [50.0, 60.0],
                    [60.0, 70.0],
                    [70.0, 80.0],
                    [80.0, 90.0],
                    [90.0, 100.0],
                    [100.0, 110.0],
                ]
            ),
            decimal=7,
        )

        multi_signals = self._multi_signals.copy()
        multi_signals.extrapolator_kwargs = {
            "method": "Linear",
        }
        np.testing.assert_array_equal(
            multi_signals[np.array([-1000, 1000])],
            np.array(
                [
                    [-9990.0, -9980.0, -9970.0],
                    [10010.0, 10020.0, 10030.0],
                ]
            ),
        )

        multi_signals.extrapolator_kwargs = {
            "method": "Constant",
            "left": 0,
            "right": 1,
        }
        np.testing.assert_array_equal(
            multi_signals[np.array([-1000, 1000])],
            np.array([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]]),
        )

    def test__setitem__(self):
        """
        Test :meth:`colour.continuous.multi_signals.MultiSignals.__setitem__`
        method.
        """

        multi_signals = self._multi_signals.copy()

        multi_signals[0] = 20
        np.testing.assert_almost_equal(
            multi_signals[0], np.array([20.0, 20.0, 20.0]), decimal=7
        )

        multi_signals[np.array([0, 1, 2])] = 30
        np.testing.assert_almost_equal(
            multi_signals[np.array([0, 1, 2])],
            np.array(
                [
                    [30.0, 30.0, 30.0],
                    [30.0, 30.0, 30.0],
                    [30.0, 30.0, 30.0],
                ]
            ),
            decimal=7,
        )

        multi_signals[np.linspace(0, 5, 5)] = 50
        np.testing.assert_almost_equal(
            multi_signals.domain,
            np.array(
                [
                    0.00,
                    1.00,
                    1.25,
                    2.00,
                    2.50,
                    3.00,
                    3.75,
                    4.00,
                    5.00,
                    6.00,
                    7.00,
                    8.00,
                    9.00,
                ]
            ),
            decimal=7,
        )
        np.testing.assert_almost_equal(
            multi_signals.range,
            np.array(
                [
                    [50.0, 50.0, 50.0],
                    [30.0, 30.0, 30.0],
                    [50.0, 50.0, 50.0],
                    [30.0, 30.0, 30.0],
                    [50.0, 50.0, 50.0],
                    [40.0, 50.0, 60.0],
                    [50.0, 50.0, 50.0],
                    [50.0, 60.0, 70.0],
                    [50.0, 50.0, 50.0],
                    [70.0, 80.0, 90.0],
                    [80.0, 90.0, 100.0],
                    [90.0, 100.0, 110.0],
                    [100.0, 110.0, 120.0],
                ]
            ),
            decimal=7,
        )

        multi_signals[np.array([0, 1, 2])] = np.array([10, 20, 30])
        np.testing.assert_almost_equal(
            multi_signals.range,
            np.array(
                [
                    [10.0, 20.0, 30.0],
                    [10.0, 20.0, 30.0],
                    [50.0, 50.0, 50.0],
                    [10.0, 20.0, 30.0],
                    [50.0, 50.0, 50.0],
                    [40.0, 50.0, 60.0],
                    [50.0, 50.0, 50.0],
                    [50.0, 60.0, 70.0],
                    [50.0, 50.0, 50.0],
                    [70.0, 80.0, 90.0],
                    [80.0, 90.0, 100.0],
                    [90.0, 100.0, 110.0],
                    [100.0, 110.0, 120.0],
                ]
            ),
            decimal=7,
        )

        multi_signals[np.array([0, 1, 2])] = np.reshape(
            np.arange(1, 10, 1), (3, 3)
        )
        np.testing.assert_almost_equal(
            multi_signals.range,
            np.array(
                [
                    [1.0, 2.0, 3.0],
                    [4.0, 5.0, 6.0],
                    [50.0, 50.0, 50.0],
                    [7.0, 8.0, 9.0],
                    [50.0, 50.0, 50.0],
                    [40.0, 50.0, 60.0],
                    [50.0, 50.0, 50.0],
                    [50.0, 60.0, 70.0],
                    [50.0, 50.0, 50.0],
                    [70.0, 80.0, 90.0],
                    [80.0, 90.0, 100.0],
                    [90.0, 100.0, 110.0],
                    [100.0, 110.0, 120.0],
                ]
            ),
            decimal=7,
        )

        multi_signals[:] = 40
        np.testing.assert_almost_equal(multi_signals.range, 40, decimal=7)

        multi_signals[:, :] = 50
        np.testing.assert_almost_equal(multi_signals.range, 50, decimal=7)

        multi_signals = self._multi_signals.copy()
        multi_signals[0:3] = 40
        np.testing.assert_almost_equal(
            multi_signals[0:3],
            np.array(
                [
                    [40.0, 40.0, 40.0],
                    [40.0, 40.0, 40.0],
                    [40.0, 40.0, 40.0],
                ]
            ),
            decimal=7,
        )

        multi_signals[:, 0:2] = 50
        np.testing.assert_almost_equal(
            multi_signals.range,
            np.array(
                [
                    [50.0, 50.0, 40.0],
                    [50.0, 50.0, 40.0],
                    [50.0, 50.0, 40.0],
                    [50.0, 50.0, 60.0],
                    [50.0, 50.0, 70.0],
                    [50.0, 50.0, 80.0],
                    [50.0, 50.0, 90.0],
                    [50.0, 50.0, 100.0],
                    [50.0, 50.0, 110.0],
                    [50.0, 50.0, 120.0],
                ]
            ),
            decimal=7,
        )

    def test__contains__(self):
        """
        Test :meth:`colour.continuous.multi_signals.MultiSignals.__contains__`
        method.
        """

        self.assertIn(0, self._multi_signals)
        self.assertIn(0.5, self._multi_signals)
        self.assertNotIn(1000, self._multi_signals)

    def test__len__(self):
        """
        Test :meth:`colour.continuous.multi_signals.MultiSignals.__len__`
        method.
        """

        self.assertEqual(len(self._multi_signals), 10)

    def test__eq__(self):
        """
        Test :meth:`colour.continuous.multi_signals.MultiSignals.__eq__`
        method.
        """

        signal_1 = self._multi_signals.copy()
        signal_2 = self._multi_signals.copy()

        self.assertEqual(signal_1, signal_2)

        self.assertNotEqual(signal_1, None)

    def test__ne__(self):
        """
        Test :meth:`colour.continuous.multi_signals.MultiSignals.__ne__`
        method.
        """

        multi_signals_1 = self._multi_signals.copy()
        multi_signals_2 = self._multi_signals.copy()

        multi_signals_2[0] = 20
        self.assertNotEqual(multi_signals_1, multi_signals_2)

        multi_signals_2[0] = np.array([10, 20, 30])
        self.assertEqual(multi_signals_1, multi_signals_2)

        multi_signals_2.interpolator = CubicSplineInterpolator
        self.assertNotEqual(multi_signals_1, multi_signals_2)

        multi_signals_2.interpolator = KernelInterpolator
        self.assertEqual(multi_signals_1, multi_signals_2)

        multi_signals_2.interpolator_kwargs = {"window": 1}
        self.assertNotEqual(multi_signals_1, multi_signals_2)

        multi_signals_2.interpolator_kwargs = {}
        self.assertEqual(multi_signals_1, multi_signals_2)

        class NotExtrapolator(Extrapolator):
            """Not :class:`Extrapolator` class."""

            pass

        multi_signals_2.extrapolator = NotExtrapolator
        self.assertNotEqual(multi_signals_1, multi_signals_2)

        multi_signals_2.extrapolator = Extrapolator
        self.assertEqual(multi_signals_1, multi_signals_2)

        multi_signals_2.extrapolator_kwargs = {}
        self.assertNotEqual(multi_signals_1, multi_signals_2)

        multi_signals_2.extrapolator_kwargs = {
            "method": "Constant",
            "left": np.nan,
            "right": np.nan,
        }
        self.assertEqual(multi_signals_1, multi_signals_2)

    def test_arithmetical_operation(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.\
arithmetical_operation` method.
        """

        np.testing.assert_almost_equal(
            self._multi_signals.arithmetical_operation(10, "+", False).range,
            self._range_2 + 10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            self._multi_signals.arithmetical_operation(10, "-", False).range,
            self._range_2 - 10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            self._multi_signals.arithmetical_operation(10, "*", False).range,
            self._range_2 * 10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            self._multi_signals.arithmetical_operation(10, "/", False).range,
            self._range_2 / 10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            self._multi_signals.arithmetical_operation(10, "**", False).range,
            self._range_2**10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            (self._multi_signals + 10).range, self._range_2 + 10, decimal=7
        )

        np.testing.assert_almost_equal(
            (self._multi_signals - 10).range, self._range_2 - 10, decimal=7
        )

        np.testing.assert_almost_equal(
            (self._multi_signals * 10).range, self._range_2 * 10, decimal=7
        )

        np.testing.assert_almost_equal(
            (self._multi_signals / 10).range, self._range_2 / 10, decimal=7
        )

        np.testing.assert_almost_equal(
            (self._multi_signals**10).range, self._range_2**10, decimal=7
        )

        multi_signals = self._multi_signals.copy()

        np.testing.assert_almost_equal(
            multi_signals.arithmetical_operation(10, "+", True).range,
            self._range_2 + 10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            multi_signals.arithmetical_operation(10, "-", True).range,
            self._range_2,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            multi_signals.arithmetical_operation(10, "*", True).range,
            self._range_2 * 10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            multi_signals.arithmetical_operation(10, "/", True).range,
            self._range_2,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            multi_signals.arithmetical_operation(10, "**", True).range,
            self._range_2**10,
            decimal=7,
        )

        multi_signals = self._multi_signals.copy()
        np.testing.assert_almost_equal(
            multi_signals.arithmetical_operation(
                self._range_2, "+", False
            ).range,
            self._range_2 + self._range_2,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            multi_signals.arithmetical_operation(
                multi_signals, "+", False
            ).range,
            self._range_2 + self._range_2,
            decimal=7,
        )

    def test_is_uniform(self):
        """
        Test :meth:`colour.continuous.multi_signals.MultiSignals.is_uniform`
        method.
        """

        self.assertTrue(self._multi_signals.is_uniform())

        multi_signals = self._multi_signals.copy()
        multi_signals[0.5] = 1.0
        self.assertFalse(multi_signals.is_uniform())

    def test_copy(self):
        """Test :func:`colour.continuous.multi_signals.MultiSignals.copy` method."""

        self.assertIsNot(self._multi_signals, self._multi_signals.copy())
        self.assertEqual(self._multi_signals, self._multi_signals.copy())

    def test_multi_signals_unpack_data(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.\
multi_signals_unpack_data` method.
        """

        signals = MultiSignals.multi_signals_unpack_data(self._range_1)
        self.assertListEqual(list(signals.keys()), ["0"])
        np.testing.assert_array_equal(signals["0"].domain, self._domain_1)
        np.testing.assert_array_equal(signals["0"].range, self._range_1)

        signals = MultiSignals.multi_signals_unpack_data(
            self._range_1, self._domain_2
        )
        self.assertListEqual(list(signals.keys()), ["0"])
        np.testing.assert_array_equal(signals["0"].domain, self._domain_2)
        np.testing.assert_array_equal(signals["0"].range, self._range_1)

        signals = MultiSignals.multi_signals_unpack_data(
            self._range_1, dict(zip(self._domain_2, self._range_1)).keys()
        )
        np.testing.assert_array_equal(signals["0"].domain, self._domain_2)

        signals = MultiSignals.multi_signals_unpack_data(
            self._range_2, self._domain_2
        )
        self.assertListEqual(list(signals.keys()), ["0", "1", "2"])
        np.testing.assert_array_equal(signals["0"].range, self._range_1)
        np.testing.assert_array_equal(signals["1"].range, self._range_1 + 10)
        np.testing.assert_array_equal(signals["2"].range, self._range_1 + 20)

        signals = MultiSignals.multi_signals_unpack_data(
            list(
                MultiSignals.multi_signals_unpack_data(
                    dict(zip(self._domain_2, self._range_2))
                ).values()
            )[0]
        )
        np.testing.assert_array_equal(signals["0"].range, self._range_1)

        signals = MultiSignals.multi_signals_unpack_data(
            MultiSignals.multi_signals_unpack_data(
                dict(zip(self._domain_2, self._range_2))
            ).values()
        )
        np.testing.assert_array_equal(signals["0"].range, self._range_1)
        np.testing.assert_array_equal(signals["1"].range, self._range_1 + 10)
        np.testing.assert_array_equal(signals["2"].range, self._range_1 + 20)

        signals = MultiSignals.multi_signals_unpack_data(
            dict(zip(self._domain_2, self._range_2))
        )
        self.assertListEqual(list(signals.keys()), ["0", "1", "2"])
        np.testing.assert_array_equal(signals["0"].range, self._range_1)
        np.testing.assert_array_equal(signals["1"].range, self._range_1 + 10)
        np.testing.assert_array_equal(signals["2"].range, self._range_1 + 20)

        signals = MultiSignals.multi_signals_unpack_data(
            MultiSignals.multi_signals_unpack_data(
                dict(zip(self._domain_2, self._range_2))
            )
        )
        self.assertListEqual(list(signals.keys()), ["0", "1", "2"])
        np.testing.assert_array_equal(signals["0"].range, self._range_1)
        np.testing.assert_array_equal(signals["1"].range, self._range_1 + 10)
        np.testing.assert_array_equal(signals["2"].range, self._range_1 + 20)

        if is_pandas_installed():
            from pandas import DataFrame, Series

            signals = MultiSignals.multi_signals_unpack_data(
                Series(dict(zip(self._domain_1, self._range_1)))
            )
            self.assertListEqual(list(signals.keys()), ["0"])
            np.testing.assert_array_equal(signals["0"].domain, self._domain_1)
            np.testing.assert_array_equal(signals["0"].range, self._range_1)

            data = dict(zip(["a", "b", "c"], tsplit(self._range_2)))
            signals = MultiSignals.multi_signals_unpack_data(
                DataFrame(data, self._domain_1)
            )
            self.assertListEqual(list(signals.keys()), ["a", "b", "c"])
            np.testing.assert_array_equal(signals["a"].range, self._range_1)
            np.testing.assert_array_equal(
                signals["b"].range, self._range_1 + 10
            )
            np.testing.assert_array_equal(
                signals["c"].range, self._range_1 + 20
            )

    def test_fill_nan(self):
        """
        Test :meth:`colour.continuous.multi_signals.MultiSignals.fill_nan`
        method.
        """

        multi_signals = self._multi_signals.copy()

        multi_signals[3:7] = np.nan

        np.testing.assert_almost_equal(
            multi_signals.fill_nan().range,
            np.array(
                [
                    [10.0, 20.0, 30.0],
                    [20.0, 30.0, 40.0],
                    [30.0, 40.0, 50.0],
                    [40.0, 50.0, 60.0],
                    [50.0, 60.0, 70.0],
                    [60.0, 70.0, 80.0],
                    [70.0, 80.0, 90.0],
                    [80.0, 90.0, 100.0],
                    [90.0, 100.0, 110.0],
                    [100.0, 110.0, 120.0],
                ]
            ),
            decimal=7,
        )

        multi_signals[3:7] = np.nan

        np.testing.assert_almost_equal(
            multi_signals.fill_nan(method="Constant").range,
            np.array(
                [
                    [10.0, 20.0, 30.0],
                    [20.0, 30.0, 40.0],
                    [30.0, 40.0, 50.0],
                    [0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0],
                    [80.0, 90.0, 100.0],
                    [90.0, 100.0, 110.0],
                    [100.0, 110.0, 120.0],
                ]
            ),
            decimal=7,
        )

    def test_domain_distance(self):
        """
        Test :func:`colour.continuous.multi_signals.MultiSignals.\
domain_distance` method.
        """

        self.assertAlmostEqual(
            self._multi_signals.domain_distance(0.5), 0.5, places=7
        )

        np.testing.assert_almost_equal(
            self._multi_signals.domain_distance(np.linspace(0, 9, 10) + 0.5),
            np.array([0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]),
            decimal=7,
        )

    def test_to_dataframe(self):
        """
        Test :meth:`colour.continuous.multi_signals.MultiSignals.to_dataframe`
        method.
        """

        if is_pandas_installed():
            from pandas import DataFrame

            data = dict(zip(["a", "b", "c"], tsplit(self._range_2)))

            attest(
                MultiSignals(
                    self._range_2, self._domain_2, labels=["a", "b", "c"]
                )
                .to_dataframe()
                .equals(DataFrame(data, self._domain_2))
            )


if __name__ == "__main__":
    unittest.main()



"""
Academy Color Encoding System - Log Encodings
=============================================

Defines the *Academy Color Encoding System* (ACES) log encodings:

-   :func:`colour.models.log_encoding_ACESproxy`
-   :func:`colour.models.log_decoding_ACESproxy`
-   :func:`colour.models.log_encoding_ACEScc`
-   :func:`colour.models.log_decoding_ACEScc`
-   :func:`colour.models.log_encoding_ACEScct`
-   :func:`colour.models.log_decoding_ACEScct`

References
----------
-   :cite:`TheAcademyofMotionPictureArtsandSciences2014q` : The Academy of
    Motion Picture Arts and Sciences, Science and Technology Council, & Academy
    Color Encoding System (ACES) Project Subcommittee. (2014). Technical
    Bulletin TB-2014-004 - Informative Notes on SMPTE ST 2065-1 - Academy Color
    Encoding Specification (ACES) (pp. 1-40). Retrieved December 19, 2014, from
    http://j.mp/TB-2014-004
-   :cite:`TheAcademyofMotionPictureArtsandSciences2014r` : The Academy of
    Motion Picture Arts and Sciences, Science and Technology Council, & Academy
    Color Encoding System (ACES) Project Subcommittee. (2014). Technical
    Bulletin TB-2014-012 - Academy Color Encoding System Version 1.0 Component
    Names (pp. 1-8). Retrieved December 19, 2014, from http://j.mp/TB-2014-012
-   :cite:`TheAcademyofMotionPictureArtsandSciences2014s` : The Academy of
    Motion Picture Arts and Sciences, Science and Technology Council, & Academy
    Color Encoding System (ACES) Project Subcommittee. (2013). Specification
    S-2013-001 - ACESproxy, an Integer Log Encoding of ACES Image Data.
    Retrieved December 19, 2014, from http://j.mp/S-2013-001
-   :cite:`TheAcademyofMotionPictureArtsandSciences2014t` : The Academy of
    Motion Picture Arts and Sciences, Science and Technology Council, & Academy
    Color Encoding System (ACES) Project Subcommittee. (2014). Specification
    S-2014-003 - ACEScc, A Logarithmic Encoding of ACES Data for use within
    Color Grading Systems (pp. 1-12). Retrieved December 19, 2014, from
    http://j.mp/S-2014-003
-   :cite:`TheAcademyofMotionPictureArtsandSciences2016c` : The Academy of
    Motion Picture Arts and Sciences, Science and Technology Council, & Academy
    Color Encoding System (ACES) Project. (2016). Specification S-2016-001 -
    ACEScct, A Quasi-Logarithmic Encoding of ACES Data for use within Color
    Grading Systems. Retrieved October 10, 2016, from http://j.mp/S-2016-001
-   :cite:`TheAcademyofMotionPictureArtsandSciencese` : The Academy of Motion
    Picture Arts and Sciences, Science and Technology Council, & Academy Color
    Encoding System (ACES) Project Subcommittee. (n.d.). Academy Color Encoding
    System. Retrieved February 24, 2014, from
    http://www.oscars.org/science-technology/council/projects/aces.html
"""

from __future__ import annotations

import numpy as np

from colour.hints import (
    Boolean,
    Dict,
    Floating,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    IntegerOrArrayLike,
    IntegerOrNDArray,
    Literal,
    Union,
)
from colour.utilities import (
    Structure,
    as_float,
    as_int,
    from_range_1,
    to_domain_1,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CONSTANTS_ACES_PROXY_10",
    "CONSTANTS_ACES_PROXY_12",
    "CONSTANTS_ACES_PROXY",
    "CONSTANTS_ACES_CCT",
    "log_encoding_ACESproxy",
    "log_decoding_ACESproxy",
    "log_encoding_ACEScc",
    "log_decoding_ACEScc",
    "log_encoding_ACEScct",
    "log_decoding_ACEScct",
]

CONSTANTS_ACES_PROXY_10: Structure = Structure(
    CV_min=64,
    CV_max=940,
    steps_per_stop=50,
    mid_CV_offset=425,
    mid_log_offset=2.5,
)
"""*ACESproxy* 10 bit colourspace constants."""

CONSTANTS_ACES_PROXY_12: Structure = Structure(
    CV_min=256,
    CV_max=3760,
    steps_per_stop=200,
    mid_CV_offset=1700,
    mid_log_offset=2.5,
)
"""*ACESproxy* 12 bit colourspace constants."""

CONSTANTS_ACES_PROXY: Dict = {
    10: CONSTANTS_ACES_PROXY_10,
    12: CONSTANTS_ACES_PROXY_12,
}
"""Aggregated *ACESproxy* colourspace constants."""

CONSTANTS_ACES_CCT: Structure = Structure(
    X_BRK=0.0078125,
    Y_BRK=0.155251141552511,
    A=10.5402377416545,
    B=0.0729055341958355,
)
"""*ACEScct* colourspace constants."""


# pylint: disable=W0102
def log_encoding_ACESproxy(
    lin_AP1: FloatingOrArrayLike,
    bit_depth: Literal[10, 12] = 10,
    out_int: Boolean = False,
    constants: Dict = CONSTANTS_ACES_PROXY,
) -> Union[FloatingOrNDArray, IntegerOrNDArray]:
    """
    Define the *ACESproxy* colourspace log encoding curve / opto-electronic
    transfer function.

    Parameters
    ----------
    lin_AP1
        *lin_AP1* value.
    bit_depth
        *ACESproxy* bit depth.
    out_in
        Whether to return value as integer code value or float equivalent of a
        code value at a given bit depth.
    constants
        *ACESproxy* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.integer` or :class:`numpy.ndarray`
        *ACESproxy* non-linear value.

    Notes
    -----
    +----------------+-----------------------+---------------+
    | **Domain \\***  | **Scale - Reference** | **Scale - 1** |
    +================+=======================+===============+
    | ``lin_AP1``    | [0, 1]                | [0, 1]        |
    +----------------+-----------------------+---------------+

    +----------------+-----------------------+---------------+
    | **Range \\***   | **Scale - Reference** | **Scale - 1** |
    +================+=======================+===============+
    | ``ACESproxy``  | [0, 1]                | [0, 1]        |
    +----------------+-----------------------+---------------+

    \\* This definition has an output integer switch, thus the domain-range
    scale information is only given for the floating point mode.

    References
    ----------
    :cite:`TheAcademyofMotionPictureArtsandSciences2014q`,
    :cite:`TheAcademyofMotionPictureArtsandSciences2014r`,
    :cite:`TheAcademyofMotionPictureArtsandSciences2014s`,
    :cite:`TheAcademyofMotionPictureArtsandSciencese`

    Examples
    --------
    >>> log_encoding_ACESproxy(0.18)  # doctest: +ELLIPSIS
    0.4164222...
    >>> log_encoding_ACESproxy(0.18, out_int=True)
    426
    """

    lin_AP1 = to_domain_1(lin_AP1)

    CV_min = constants[bit_depth].CV_min
    CV_max = constants[bit_depth].CV_max
    mid_CV_offset = constants[bit_depth].mid_CV_offset
    mid_log_offset = constants[bit_depth].mid_log_offset
    steps_per_stop = constants[bit_depth].steps_per_stop

    def float_2_cv(x: Floating) -> Floating:
        """Convert given numeric to code value."""

        return np.maximum(CV_min, np.minimum(CV_max, np.round(x)))

    ACESproxy = np.where(
        lin_AP1 > 2**-9.72,
        float_2_cv(
            (np.log2(lin_AP1) + mid_log_offset) * steps_per_stop
            + mid_CV_offset
        ),
        np.resize(CV_min, lin_AP1.shape),
    )

    if out_int:
        return as_int(np.round(ACESproxy))
    else:
        return as_float(from_range_1(ACESproxy / (2**bit_depth - 1)))


# pylint: disable=W0102
def log_decoding_ACESproxy(
    ACESproxy: Union[FloatingOrArrayLike, IntegerOrArrayLike],
    bit_depth: Literal[10, 12] = 10,
    in_int: Boolean = False,
    constants: Dict = CONSTANTS_ACES_PROXY,
) -> FloatingOrNDArray:
    """
    Define the *ACESproxy* colourspace log decoding curve / electro-optical
    transfer function.

    Parameters
    ----------
    ACESproxy
        *ACESproxy* non-linear value.
    bit_depth
        *ACESproxy* bit depth.
    in_int
        Whether to treat the input value as integer code value or float
        equivalent of a code value at a given bit depth.
    constants
        *ACESproxy* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *lin_AP1* value.

    Notes
    -----
    +----------------+-----------------------+---------------+
    | **Domain \\***  | **Scale - Reference** | **Scale - 1** |
    +================+=======================+===============+
    | ``ACESproxy``  | [0, 1]                | [0, 1]        |
    +----------------+-----------------------+---------------+

    +----------------+-----------------------+---------------+
    | **Range \\***   | **Scale - Reference** | **Scale - 1** |
    +================+=======================+===============+
    | ``lin_AP1``    | [0, 1]                | [0, 1]        |
    +----------------+-----------------------+---------------+

    \\* This definition has an input integer switch, thus the domain-range
    scale information is only given for the floating point mode.

    References
    ----------
    :cite:`TheAcademyofMotionPictureArtsandSciences2014q`,
    :cite:`TheAcademyofMotionPictureArtsandSciences2014r`,
    :cite:`TheAcademyofMotionPictureArtsandSciences2014s`,
    :cite:`TheAcademyofMotionPictureArtsandSciencese`

    Examples
    --------
    >>> log_decoding_ACESproxy(0.416422287390029)  # doctest: +ELLIPSIS
    0.1...
    >>> log_decoding_ACESproxy(426, in_int=True)  # doctest: +ELLIPSIS
    0.1...
    """

    ACESproxy = to_domain_1(ACESproxy)

    mid_CV_offset = constants[bit_depth].mid_CV_offset
    mid_log_offset = constants[bit_depth].mid_log_offset
    steps_per_stop = constants[bit_depth].steps_per_stop

    if not in_int:
        ACESproxy = ACESproxy * (2**bit_depth - 1)

    lin_AP1 = 2 ** (
        (ACESproxy - mid_CV_offset) / steps_per_stop - mid_log_offset
    )

    return as_float(from_range_1(lin_AP1))


def log_encoding_ACEScc(lin_AP1: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define the *ACEScc* colourspace log encoding / opto-electronic transfer
    function.

    Parameters
    ----------
    lin_AP1
        *lin_AP1* value.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *ACEScc* non-linear value.

    Notes
    -----
    +-------------+-----------------------+---------------+
    | **Domain**  | **Scale - Reference** | **Scale - 1** |
    +=============+=======================+===============+
    | ``lin_AP1`` | [0, 1]                | [0, 1]        |
    +-------------+-----------------------+---------------+

    +-------------+-----------------------+---------------+
    | **Range**   | **Scale - Reference** | **Scale - 1** |
    +=============+=======================+===============+
    | ``ACEScc``  | [0, 1]                | [0, 1]        |
    +-------------+-----------------------+---------------+

    References
    ----------
    :cite:`TheAcademyofMotionPictureArtsandSciences2014q`,
    :cite:`TheAcademyofMotionPictureArtsandSciences2014r`,
    :cite:`TheAcademyofMotionPictureArtsandSciences2014t`,
    :cite:`TheAcademyofMotionPictureArtsandSciencese`

    Examples
    --------
    >>> log_encoding_ACEScc(0.18)  # doctest: +ELLIPSIS
    0.4135884...
    """

    lin_AP1 = to_domain_1(lin_AP1)

    ACEScc = np.where(
        lin_AP1 < 0,
        (np.log2(2**-16) + 9.72) / 17.52,
        (np.log2(2**-16 + lin_AP1 * 0.5) + 9.72) / 17.52,
    )
    ACEScc = np.where(
        lin_AP1 >= 2**-15,
        (np.log2(lin_AP1) + 9.72) / 17.52,
        ACEScc,
    )

    return as_float(from_range_1(ACEScc))


def log_decoding_ACEScc(ACEScc: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define the *ACEScc* colourspace log decoding / electro-optical transfer
    function.

    Parameters
    ----------
    ACEScc
        *ACEScc* non-linear value.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *lin_AP1* value.

    Notes
    -----
    +-------------+-----------------------+---------------+
    | **Domain**  | **Scale - Reference** | **Scale - 1** |
    +=============+=======================+===============+
    | ``ACEScc``  | [0, 1]                | [0, 1]        |
    +-------------+-----------------------+---------------+

    +-------------+-----------------------+---------------+
    | **Range**   | **Scale - Reference** | **Scale - 1** |
    +=============+=======================+===============+
    | ``lin_AP1`` | [0, 1]                | [0, 1]        |
    +-------------+-----------------------+---------------+

    References
    ----------
    :cite:`TheAcademyofMotionPictureArtsandSciences2014q`,
    :cite:`TheAcademyofMotionPictureArtsandSciences2014r`,
    :cite:`TheAcademyofMotionPictureArtsandSciences2014t`,
    :cite:`TheAcademyofMotionPictureArtsandSciencese`

    Examples
    --------
    >>> log_decoding_ACEScc(0.413588402492442)  # doctest: +ELLIPSIS
    0.1799999...
    """

    ACEScc = to_domain_1(ACEScc)

    lin_AP1 = np.where(
        ACEScc < (9.72 - 15) / 17.52,
        (2 ** (ACEScc * 17.52 - 9.72) - 2**-16) * 2,
        2 ** (ACEScc * 17.52 - 9.72),
    )
    lin_AP1 = np.where(
        ACEScc >= (np.log2(65504) + 9.72) / 17.52,
        65504,
        lin_AP1,
    )

    return as_float(from_range_1(lin_AP1))


# pylint: disable=W0102
def log_encoding_ACEScct(
    lin_AP1: FloatingOrArrayLike, constants: Structure = CONSTANTS_ACES_CCT
) -> FloatingOrNDArray:
    """
    Define the *ACEScct* colourspace log encoding / opto-electronic transfer
    function.

    Parameters
    ----------
    lin_AP1
        *lin_AP1* value.
    constants
        *ACEScct* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *ACEScct* non-linear value.

    Notes
    -----
    +-------------+-----------------------+---------------+
    | **Domain**  | **Scale - Reference** | **Scale - 1** |
    +=============+=======================+===============+
    | ``lin_AP1`` | [0, 1]                | [0, 1]        |
    +-------------+-----------------------+---------------+

    +-------------+-----------------------+---------------+
    | **Range**   | **Scale - Reference** | **Scale - 1** |
    +=============+=======================+===============+
    | ``ACEScct`` | [0, 1]                | [0, 1]        |
    +-------------+-----------------------+---------------+

    References
    ----------
    :cite:`TheAcademyofMotionPictureArtsandSciences2014q`,
    :cite:`TheAcademyofMotionPictureArtsandSciences2014r`,
    :cite:`TheAcademyofMotionPictureArtsandSciences2016c`,
    :cite:`TheAcademyofMotionPictureArtsandSciencese`

    Examples
    --------
    >>> log_encoding_ACEScct(0.18)  # doctest: +ELLIPSIS
    0.4135884...
    """

    lin_AP1 = to_domain_1(lin_AP1)

    ACEScct = np.where(
        lin_AP1 <= constants.X_BRK,
        constants.A * lin_AP1 + constants.B,
        (np.log2(lin_AP1) + 9.72) / 17.52,
    )

    return as_float(from_range_1(ACEScct))


# pylint: disable=W0102
def log_decoding_ACEScct(
    ACEScct: FloatingOrArrayLike, constants: Structure = CONSTANTS_ACES_CCT
) -> FloatingOrNDArray:
    """
    Define the *ACEScct* colourspace log decoding / electro-optical transfer
    function.

    Parameters
    ----------
    ACEScct
        *ACEScct* non-linear value.
    constants
        *ACEScct* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *lin_AP1* value.

    References
    ----------
    :cite:`TheAcademyofMotionPictureArtsandSciences2014q`,
    :cite:`TheAcademyofMotionPictureArtsandSciences2014r`,
    :cite:`TheAcademyofMotionPictureArtsandSciences2016c`,
    :cite:`TheAcademyofMotionPictureArtsandSciencese`

    Notes
    -----
    +-------------+-----------------------+---------------+
    | **Domain**  | **Scale - Reference** | **Scale - 1** |
    +=============+=======================+===============+
    | ``ACEScct`` | [0, 1]                | [0, 1]        |
    +-------------+-----------------------+---------------+

    +-------------+-----------------------+---------------+
    | **Range**   | **Scale - Reference** | **Scale - 1** |
    +=============+=======================+===============+
    | ``lin_AP1`` | [0, 1]                | [0, 1]        |
    +-------------+-----------------------+---------------+

    Examples
    --------
    >>> log_decoding_ACEScct(0.413588402492442)  # doctest: +ELLIPSIS
    0.1799999...
    """

    ACEScct = to_domain_1(ACEScct)

    lin_AP1 = np.where(
        ACEScct > constants.Y_BRK,
        2 ** (ACEScct * 17.52 - 9.72),
        (ACEScct - constants.B) / constants.A,
    )

    return as_float(from_range_1(lin_AP1))



"""Defines the unit tests for the :mod:`colour.plotting.phenomena` module."""

import unittest
from matplotlib.pyplot import Axes, Figure

from colour.plotting import (
    plot_single_sd_rayleigh_scattering,
    plot_the_blue_sky,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestPlotSingleSdRayleighScattering",
    "TestPlotTheBlueSky",
]


class TestPlotSingleSdRayleighScattering(unittest.TestCase):
    """
    Define :func:`colour.plotting.phenomena.\
plot_single_sd_rayleigh_scattering` definition unit tests methods.
    """

    def test_plot_single_sd_rayleigh_scattering(self):
        """
        Test :func:`colour.plotting.phenomena.\
plot_single_sd_rayleigh_scattering` definition.
        """

        figure, axes = plot_single_sd_rayleigh_scattering()

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotTheBlueSky(unittest.TestCase):
    """
    Define :func:`colour.plotting.phenomena.plot_the_blue_sky` definition unit
    tests methods.
    """

    def test_plot_the_blue_sky(self):
        """Test :func:`colour.plotting.phenomena.plot_the_blue_sky` definition."""

        figure, axes = plot_the_blue_sky()

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


if __name__ == "__main__":
    unittest.main()



"""Defines the unit tests for the :mod:`colour.plotting.colorimetry` module."""

import unittest
from matplotlib.pyplot import Axes, Figure

from colour.colorimetry import SpectralDistribution
from colour.plotting import (
    plot_single_sd,
    plot_multi_sds,
    plot_single_cmfs,
    plot_multi_cmfs,
    plot_single_illuminant_sd,
    plot_multi_illuminant_sds,
    plot_visible_spectrum,
    plot_single_lightness_function,
    plot_multi_lightness_functions,
    plot_single_luminance_function,
    plot_multi_luminance_functions,
    plot_blackbody_spectral_radiance,
    plot_blackbody_colours,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestPlotSingleSd",
    "TestPlotMultiSds",
    "TestPlotSingleCmfs",
    "TestPlotMultiCmfs",
    "TestPlotSingleIlluminantSd",
    "TestPlotMultiIlluminantSds",
    "TestPlotVisibleSpectrum",
    "TestPlotSingleLightnessFunction",
    "TestPlotMultiLightnessFunctions",
    "TestPlotSingleLuminanceFunction",
    "TestPlotMultiLuminanceFunctions",
    "TestPlotBlackbodySpectralRadiance",
    "TestPlotBlackbodyColours",
]


class TestPlotSingleSd(unittest.TestCase):
    """
    Define :func:`colour.plotting.colorimetry.plot_single_sd` definition unit
    tests methods.
    """

    def test_plot_single_sd(self):
        """Test :func:`colour.plotting.colorimetry.plot_single_sd` definition."""

        sd = SpectralDistribution(
            {
                500: 0.004900,
                510: 0.009300,
                520: 0.063270,
                530: 0.165500,
                540: 0.290400,
                550: 0.433450,
                560: 0.594500,
            },
            name="Custom 1",
        )

        figure, axes = plot_single_sd(
            sd,
            out_of_gamut_clipping=False,
            modulate_colours_with_sd_amplitude=True,
            equalize_sd_amplitude=True,
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotMultiSds(unittest.TestCase):
    """
    Define :func:`colour.plotting.colorimetry.plot_multi_sds` definition unit
    tests methods.
    """

    def test_plot_multi_sds(self):
        """Test :func:`colour.plotting.colorimetry.plot_multi_sds` definition."""

        sd_1 = SpectralDistribution(
            {
                500: 0.004900,
                510: 0.009300,
                520: 0.063270,
                530: 0.165500,
                540: 0.290400,
                550: 0.433450,
                560: 0.594500,
            },
            name="Custom 1",
        )
        sd_2 = SpectralDistribution(
            {
                500: 0.323000,
                510: 0.503000,
                520: 0.710000,
                530: 0.862000,
                540: 0.954000,
                550: 0.994950,
                560: 0.995000,
            },
            name="Custom 2",
        )

        figure, axes = plot_multi_sds(
            [sd_1, sd_2],
            plot_kwargs={"use_sd_colours": True, "normalise_sd_colours": True},
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)

        figure, axes = plot_multi_sds(
            [sd_1, sd_2],
            plot_kwargs=[
                {"use_sd_colours": True, "normalise_sd_colours": True}
            ]
            * 2,
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotSingleCmfs(unittest.TestCase):
    """
    Define :func:`colour.plotting.colorimetry.plot_single_cmfs` definition
    unit tests methods.
    """

    def test_plot_single_cmfs(self):
        """Test :func:`colour.plotting.colorimetry.plot_single_cmfs` definition."""

        figure, axes = plot_single_cmfs()

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotMultiCmfs(unittest.TestCase):
    """
    Define :func:`colour.plotting.colorimetry.plot_multi_cmfs` definition unit
    tests methods.
    """

    def test_plot_multi_cmfs(self):
        """Test :func:`colour.plotting.colorimetry.plot_multi_cmfs` definition."""

        figure, axes = plot_multi_cmfs(
            [
                "CIE 1931 2 Degree Standard Observer",
                "CIE 1964 10 Degree Standard Observer",
            ]
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotSingleIlluminantSd(unittest.TestCase):
    """
    Define :func:`colour.plotting.colorimetry.plot_single_illuminant_sd`
    definition unit tests methods.
    """

    def test_plot_single_illuminant_sd(self):
        """
        Test :func:`colour.plotting.colorimetry.plot_single_illuminant_sd`
        definition.
        """

        figure, axes = plot_single_illuminant_sd("A")

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotMultiIlluminantSds(unittest.TestCase):
    """
    Define :func:`colour.plotting.colorimetry.plot_multi_illuminant_sds`
    definition unit tests methods.
    """

    def test_plot_multi_illuminant_sds(self):
        """
        Test :func:`colour.plotting.colorimetry.plot_multi_illuminant_sds`
        definition.
        """

        figure, axes = plot_multi_illuminant_sds(["A", "B", "C"])

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)

        figure, axes = plot_multi_illuminant_sds(
            ["A", "B", "C"],
            plot_kwargs=[
                {"use_sd_colours": True, "normalise_sd_colours": True}
            ]
            * 3,
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotVisibleSpectrum(unittest.TestCase):
    """
    Define :func:`colour.plotting.colorimetry.plot_visible_spectrum`
    definition unit tests methods.
    """

    def test_plot_visible_spectrum(self):
        """
        Test :func:`colour.plotting.colorimetry.plot_visible_spectrum`
        definition.
        """

        figure, axes = plot_visible_spectrum()

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotSingleLightnessFunction(unittest.TestCase):
    """
    Define :func:`colour.plotting.colorimetry.plot_single_lightness_function`
    definition unit tests methods.
    """

    def test_plot_single_lightness_function(self):
        """
        Test :func:`colour.plotting.colorimetry.\
plot_single_lightness_function` definition.
        """

        figure, axes = plot_single_lightness_function("CIE 1976")

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotMultiLightnessFunctions(unittest.TestCase):
    """
    Define :func:`colour.plotting.colorimetry.plot_multi_lightness_functions`
    definition unit tests methods.
    """

    def test_plot_multi_lightness_functions(self):
        """
        Test :func:`colour.plotting.colorimetry.\
plot_multi_lightness_functions` definition.
        """

        figure, axes = plot_multi_lightness_functions(
            ["CIE 1976", "Wyszecki 1963"]
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotSingleLuminanceFunction(unittest.TestCase):
    """
    Define :func:`colour.plotting.colorimetry.plot_single_luminance_function`
    definition unit tests methods.
    """

    def test_plot_single_luminance_function(self):
        """
        Test :func:`colour.plotting.colorimetry.\
plot_single_luminance_function` definition.
        """

        figure, axes = plot_single_luminance_function("CIE 1976")

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotMultiLuminanceFunctions(unittest.TestCase):
    """
    Define :func:`colour.plotting.colorimetry.plot_multi_luminance_functions`
    definition unit tests methods.
    """

    def test_plot_multi_luminance_functions(self):
        """
        Test :func:`colour.plotting.colorimetry.\
plot_multi_luminance_functions` definition.
        """

        figure, axes = plot_multi_luminance_functions(
            ["CIE 1976", "Newhall 1943"]
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotBlackbodySpectralRadiance(unittest.TestCase):
    """
    Define :func:`colour.plotting.colorimetry.\
plot_blackbody_spectral_radiance` definition unit tests methods.
    """

    def test_plot_blackbody_spectral_radiance(self):
        """
        Test :func:`colour.plotting.colorimetry.\
plot_blackbody_spectral_radiance` definition.
        """

        figure, axes = plot_blackbody_spectral_radiance()

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotBlackbodyColours(unittest.TestCase):
    """
    Define :func:`colour.plotting.colorimetry.plot_blackbody_colours`
    definition unit tests methods.
    """

    def test_plot_blackbody_colours(self):
        """
        Test :func:`colour.plotting.colorimetry.plot_blackbody_colours`
        definition.
        """

        figure, axes = plot_blackbody_colours()

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


if __name__ == "__main__":
    unittest.main()



"""Defines the unit tests for the :mod:`colour.algebra.common` module."""

import numpy as np
import unittest

from colour.algebra import (
    is_spow_enabled,
    set_spow_enable,
    spow_enable,
    spow,
    smoothstep_function,
    normalise_maximum,
    vector_dot,
    matrix_dot,
    linear_conversion,
    linstep_function,
    is_identity,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestIsSpowEnabled",
    "TestSetSpowEnabled",
    "TestSpowEnable",
    "TestSpow",
    "TestSmoothstepFunction",
    "TestNormaliseMaximum",
    "TestVectorDot",
    "TestMatrixDot",
    "TestLinearConversion",
    "TestLinstepFunction",
    "TestIsIdentity",
]


class TestIsSpowEnabled(unittest.TestCase):
    """
    Define :func:`colour.algebra.common.is_spow_enabled` definition unit
    tests methods.
    """

    def test_is_spow_enabled(self):
        """Test :func:`colour.algebra.common.is_spow_enabled` definition."""

        with spow_enable(True):
            self.assertTrue(is_spow_enabled())

        with spow_enable(False):
            self.assertFalse(is_spow_enabled())


class TestSetSpowEnabled(unittest.TestCase):
    """
    Define :func:`colour.algebra.common.set_spow_enable` definition unit
    tests methods.
    """

    def test_set_spow_enable(self):
        """Test :func:`colour.algebra.common.set_spow_enable` definition."""

        with spow_enable(is_spow_enabled()):
            set_spow_enable(True)
            self.assertTrue(is_spow_enabled())

        with spow_enable(is_spow_enabled()):
            set_spow_enable(False)
            self.assertFalse(is_spow_enabled())


class TestSpowEnable(unittest.TestCase):
    """
    Define :func:`colour.algebra.common.spow_enable` definition unit
    tests methods.
    """

    def test_spow_enable(self):
        """Test :func:`colour.algebra.common.spow_enable` definition."""

        with spow_enable(True):
            self.assertTrue(is_spow_enabled())

        with spow_enable(False):
            self.assertFalse(is_spow_enabled())

        @spow_enable(True)
        def fn_a():
            """:func:`spow_enable` unit tests :func:`fn_a` definition."""

            self.assertTrue(is_spow_enabled())

        fn_a()

        @spow_enable(False)
        def fn_b():
            """:func:`spow_enable` unit tests :func:`fn_b` definition."""

            self.assertFalse(is_spow_enabled())

        fn_b()


class TestSpow(unittest.TestCase):
    """
    Define :func:`colour.algebra.common.spow` definition unit
    tests methods.
    """

    def test_spow(self):
        """Test :func:`colour.algebra.common.spow` definition."""

        self.assertEqual(spow(2, 2), 4.0)

        self.assertEqual(spow(-2, 2), -4.0)

        np.testing.assert_almost_equal(
            spow([2, -2, -2, 0], [2, 2, 0.15, 0]),
            np.array([4.00000000, -4.00000000, -1.10956947, 0.00000000]),
            decimal=7,
        )

        with spow_enable(True):
            np.testing.assert_almost_equal(
                spow(-2, 0.15), -1.10956947, decimal=7
            )

        with spow_enable(False):
            np.testing.assert_equal(spow(-2, 0.15), np.nan)


class TestNormaliseMaximum(unittest.TestCase):
    """
    Define :func:`colour.utilities.array.normalise_maximum` definition unit
    tests methods.
    """

    def test_normalise_maximum(self):
        """Test :func:`colour.utilities.array.normalise_maximum` definition."""

        np.testing.assert_almost_equal(
            normalise_maximum(np.array([0.20654008, 0.12197225, 0.05136952])),
            np.array([1.00000000, 0.59055003, 0.24871454]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            normalise_maximum(
                np.array(
                    [
                        [0.20654008, 0.12197225, 0.05136952],
                        [0.14222010, 0.23042768, 0.10495772],
                        [0.07818780, 0.06157201, 0.28099326],
                    ]
                )
            ),
            np.array(
                [
                    [0.73503571, 0.43407536, 0.18281406],
                    [0.50613349, 0.82004700, 0.37352398],
                    [0.27825507, 0.21912273, 1.00000000],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            normalise_maximum(
                np.array(
                    [
                        [0.20654008, 0.12197225, 0.05136952],
                        [0.14222010, 0.23042768, 0.10495772],
                        [0.07818780, 0.06157201, 0.28099326],
                    ]
                ),
                axis=-1,
            ),
            np.array(
                [
                    [1.00000000, 0.59055003, 0.24871454],
                    [0.61720059, 1.00000000, 0.45549094],
                    [0.27825507, 0.21912273, 1.00000000],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            normalise_maximum(
                np.array([0.20654008, 0.12197225, 0.05136952]), factor=10
            ),
            np.array([10.00000000, 5.90550028, 2.48714535]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            normalise_maximum(
                np.array([-0.11518475, -0.10080000, 0.05089373])
            ),
            np.array([0.00000000, 0.00000000, 1.00000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            normalise_maximum(
                np.array([-0.20654008, -0.12197225, 0.05136952]), clip=False
            ),
            np.array([-4.02067374, -2.37440899, 1.00000000]),
            decimal=7,
        )


class TestVectorDot(unittest.TestCase):
    """
    Define :func:`colour.utilities.array.vector_dot` definition unit tests
    methods.
    """

    def test_vector_dot(self):
        """Test :func:`colour.utilities.array.vector_dot` definition."""

        m = np.array(
            [
                [0.7328, 0.4296, -0.1624],
                [-0.7036, 1.6975, 0.0061],
                [0.0030, 0.0136, 0.9834],
            ]
        )
        m = np.reshape(np.tile(m, (6, 1)), (6, 3, 3))

        v = np.array([0.20654008, 0.12197225, 0.05136952])
        v = np.tile(v, (6, 1))

        np.testing.assert_almost_equal(
            vector_dot(m, v),
            np.array(
                [
                    [0.19540944, 0.06203965, 0.05279523],
                    [0.19540944, 0.06203965, 0.05279523],
                    [0.19540944, 0.06203965, 0.05279523],
                    [0.19540944, 0.06203965, 0.05279523],
                    [0.19540944, 0.06203965, 0.05279523],
                    [0.19540944, 0.06203965, 0.05279523],
                ]
            ),
            decimal=7,
        )


class TestMatrixDot(unittest.TestCase):
    """
    Define :func:`colour.utilities.array.matrix_dot` definition unit tests
    methods.
    """

    def test_matrix_dot(self):
        """Test :func:`colour.utilities.array.matrix_dot` definition."""

        a = np.array(
            [
                [0.7328, 0.4296, -0.1624],
                [-0.7036, 1.6975, 0.0061],
                [0.0030, 0.0136, 0.9834],
            ]
        )
        a = np.reshape(np.tile(a, (6, 1)), (6, 3, 3))

        b = a

        np.testing.assert_almost_equal(
            matrix_dot(a, b),
            np.array(
                [
                    [
                        [0.23424208, 1.04184824, -0.27609032],
                        [-1.70994078, 2.57932265, 0.13061813],
                        [-0.00442036, 0.03774904, 0.96667132],
                    ],
                    [
                        [0.23424208, 1.04184824, -0.27609032],
                        [-1.70994078, 2.57932265, 0.13061813],
                        [-0.00442036, 0.03774904, 0.96667132],
                    ],
                    [
                        [0.23424208, 1.04184824, -0.27609032],
                        [-1.70994078, 2.57932265, 0.13061813],
                        [-0.00442036, 0.03774904, 0.96667132],
                    ],
                    [
                        [0.23424208, 1.04184824, -0.27609032],
                        [-1.70994078, 2.57932265, 0.13061813],
                        [-0.00442036, 0.03774904, 0.96667132],
                    ],
                    [
                        [0.23424208, 1.04184824, -0.27609032],
                        [-1.70994078, 2.57932265, 0.13061813],
                        [-0.00442036, 0.03774904, 0.96667132],
                    ],
                    [
                        [0.23424208, 1.04184824, -0.27609032],
                        [-1.70994078, 2.57932265, 0.13061813],
                        [-0.00442036, 0.03774904, 0.96667132],
                    ],
                ]
            ),
            decimal=7,
        )


class TestLinearConversion(unittest.TestCase):
    """
    Define :func:`colour.utilities.array.linear_conversion` definition unit
    tests methods.
    """

    def test_linear_conversion(self):
        """Test :func:`colour.utilities.array.linear_conversion` definition."""

        np.testing.assert_almost_equal(
            linear_conversion(
                np.linspace(0, 1, 10), np.array([0, 1]), np.array([1, np.pi])
            ),
            np.array(
                [
                    1.00000000,
                    1.23795474,
                    1.47590948,
                    1.71386422,
                    1.95181896,
                    2.18977370,
                    2.42772844,
                    2.66568318,
                    2.90363791,
                    3.14159265,
                ]
            ),
            decimal=8,
        )


class TestLinstepFunction(unittest.TestCase):
    """
    Define :func:`colour.utilities.array.linstep_function` definition unit
    tests methods.
    """

    def test_linstep_function(self):
        """Test :func:`colour.utilities.array.linstep_function` definition."""

        np.testing.assert_almost_equal(
            linstep_function(
                np.linspace(0, 1, 10),
                np.linspace(0, 1, 10),
                np.linspace(0, 2, 10),
            ),
            np.array(
                [
                    0.00000000,
                    0.12345679,
                    0.27160494,
                    0.44444444,
                    0.64197531,
                    0.86419753,
                    1.11111111,
                    1.38271605,
                    1.67901235,
                    2.00000000,
                ]
            ),
            decimal=8,
        )

        np.testing.assert_almost_equal(
            linstep_function(
                np.linspace(0, 2, 10),
                np.linspace(0.25, 0.5, 10),
                np.linspace(0.5, 0.75, 10),
                clip=True,
            ),
            np.array(
                [
                    0.25000000,
                    0.33333333,
                    0.41666667,
                    0.50000000,
                    0.58333333,
                    0.63888889,
                    0.66666667,
                    0.69444444,
                    0.72222222,
                    0.75000000,
                ]
            ),
            decimal=8,
        )


class TestSmoothstepFunction(unittest.TestCase):
    """
    Define :func:`colour.algebra.common.smoothstep_function` definition unit
    tests methods.
    """

    def test_smoothstep_function(self):
        """Test :func:`colour.algebra.common.smoothstep_function` definition."""

        self.assertEqual(smoothstep_function(0.5), 0.5)
        self.assertEqual(smoothstep_function(0.25), 0.15625)
        self.assertEqual(smoothstep_function(0.75), 0.84375)

        x = np.linspace(-2, 2, 5)
        np.testing.assert_almost_equal(
            smoothstep_function(x),
            np.array([28.00000, 5.00000, 0.00000, 1.00000, -4.00000]),
        )
        np.testing.assert_almost_equal(
            smoothstep_function(x, -2, 2, clip=True),
            np.array([0.00000, 0.15625, 0.50000, 0.84375, 1.00000]),
        )


class TestIsIdentity(unittest.TestCase):
    """
    Define :func:`colour.algebra.matrix.is_identity` definition unit tests
    methods.
    """

    def test_is_identity(self):
        """Test :func:`colour.algebra.matrix.is_identity` definition."""

        self.assertTrue(
            is_identity(np.array([1, 0, 0, 0, 1, 0, 0, 0, 1]).reshape([3, 3]))
        )

        self.assertFalse(
            is_identity(np.array([1, 2, 0, 0, 1, 0, 0, 0, 1]).reshape([3, 3]))
        )

        self.assertTrue(is_identity(np.array([1, 0, 0, 1]).reshape([2, 2])))

        self.assertFalse(is_identity(np.array([1, 2, 0, 1]).reshape([2, 2])))


if __name__ == "__main__":
    unittest.main()



"""Defines the unit tests for the :mod:`colour.recovery.meng2015` module."""

import numpy as np
import unittest

from colour.colorimetry import (
    MSDS_CMFS,
    SDS_ILLUMINANTS,
    SpectralShape,
    reshape_msds,
    reshape_sd,
    sd_to_XYZ_integration,
)
from colour.recovery import XYZ_to_sd_Meng2015
from colour.utilities import domain_range_scale

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestXYZ_to_sd_Meng2015",
]


class TestXYZ_to_sd_Meng2015(unittest.TestCase):
    """
    Define :func:`colour.recovery.meng2015.XYZ_to_sd_Meng2015` definition unit
    tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        # pylint: disable=E1102
        self._cmfs = reshape_msds(
            MSDS_CMFS["CIE 1931 2 Degree Standard Observer"],
            SpectralShape(360, 780, 10),
        )
        self._sd_D65 = reshape_sd(SDS_ILLUMINANTS["D65"], self._cmfs.shape)
        self._sd_E = reshape_sd(SDS_ILLUMINANTS["E"], self._cmfs.shape)

    def test_XYZ_to_sd_Meng2015(self):
        """Test :func:`colour.recovery.meng2015.XYZ_to_sd_Meng2015` definition."""

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
        np.testing.assert_almost_equal(
            sd_to_XYZ_integration(
                XYZ_to_sd_Meng2015(XYZ, self._cmfs, self._sd_D65),
                self._cmfs,
                self._sd_D65,
            )
            / 100,
            XYZ,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            sd_to_XYZ_integration(
                XYZ_to_sd_Meng2015(XYZ, self._cmfs, self._sd_E),
                self._cmfs,
                self._sd_E,
            )
            / 100,
            XYZ,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            sd_to_XYZ_integration(
                XYZ_to_sd_Meng2015(
                    XYZ,
                    self._cmfs,
                    self._sd_D65,
                    optimisation_kwargs={
                        "options": {
                            "ftol": 1e-10,
                        }
                    },
                ),
                self._cmfs,
                self._sd_D65,
            )
            / 100,
            XYZ,
            decimal=7,
        )

        shape = SpectralShape(400, 700, 5)
        # pylint: disable=E1102
        cmfs = reshape_msds(self._cmfs, shape)
        np.testing.assert_almost_equal(
            sd_to_XYZ_integration(
                XYZ_to_sd_Meng2015(XYZ, cmfs, self._sd_D65), cmfs, self._sd_D65
            )
            / 100,
            XYZ,
            decimal=7,
        )

    def test_raise_exception_XYZ_to_sd_Meng2015(self):
        """
        Test :func:`colour.recovery.meng2015.XYZ_to_sd_Meng2015`
        definition raised exception.
        """

        self.assertRaises(
            RuntimeError,
            XYZ_to_sd_Meng2015,
            np.array([0.0, 0.0, 1.0]),
            optimisation_kwargs={
                "options": {"maxiter": 10},
            },
        )

    def test_domain_range_scale_XYZ_to_sd_Meng2015(self):
        """
        Test :func:`colour.recovery.meng2015.XYZ_to_sd_Meng2015` definition
        domain and range scale support.
        """

        XYZ_i = np.array([0.20654008, 0.12197225, 0.05136952])
        XYZ_o = sd_to_XYZ_integration(
            XYZ_to_sd_Meng2015(XYZ_i, self._cmfs, self._sd_D65),
            self._cmfs,
            self._sd_D65,
        )

        d_r = (("reference", 1, 1), ("1", 1, 0.01), ("100", 100, 1))
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    sd_to_XYZ_integration(
                        XYZ_to_sd_Meng2015(
                            XYZ_i * factor_a, self._cmfs, self._sd_D65
                        ),
                        self._cmfs,
                        self._sd_D65,
                    ),
                    XYZ_o * factor_b,
                    decimal=7,
                )


if __name__ == "__main__":
    unittest.main()



"""Showcases *Zhai and Luo (2018)* chromatic adaptation model computations."""

import numpy as np

import colour
from colour.utilities import message_box

message_box('"Zhai and Luo (2018)" Chromatic Adaptation Model Computations')

XYZ_b = np.array([0.48900, 0.43620, 0.06250])
XYZ_wb = np.array([1.09850, 1, 0.35585])
XYZ_wd = np.array([0.95047, 1, 1.08883])
D_b = 0.9407
D_d = 0.9800
XYZ_wo = np.array([1, 1, 1])
message_box(
    f'Computing chromatic adaptation using "Zhai and Luo (2018)" chromatic '
    f"adaptation model.\n\n"
    f'\t"XYZ_b": {XYZ_b}\n'
    f'\t"XYZ_wb": {XYZ_wb}\n'
    f'\t"XYZ_wd": {XYZ_wd}\n'
    f'\t"D_b": {D_b}\n'
    f'\t"D_d": {D_d}\n'
    f'\t"XYZ_wo": {XYZ_wo}'
)
print(
    colour.chromatic_adaptation(
        XYZ_b,
        XYZ_wb,
        XYZ_wd,
        method="Zhai 2018",
        D_b=D_b,
        D_d=D_d,
        XYZ_wo=XYZ_wo,
    )
)
print(
    colour.adaptation.chromatic_adaptation_Zhai2018(
        XYZ_b * 100, XYZ_wb * 100, XYZ_wd * 100, D_b, D_d, XYZ_wo * 100
    )
    / 100
)



"""
Defines the unit tests for the :mod:`colour.algebra.random` module.

References
----------
-   :cite:`Laurent2012a` : Laurent. (2012). Reproducibility of python
    pseudo-random numbers across systems and versions? Retrieved January 20,
    2015, from
    http://stackoverflow.com/questions/8786084/\
reproducibility-of-python-pseudo-random-numbers-across-systems-and-versions
"""

from __future__ import annotations

import numpy as np
import unittest

from colour.algebra import random_triplet_generator
from colour.hints import NDArray

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "RANDOM_TRIPLETS",
    "TestRandomTripletGenerator",
]

RANDOM_TRIPLETS: NDArray = np.array(
    [
        [0.96702984, 0.77938292, 0.43614665],
        [0.54723225, 0.19768507, 0.94897731],
        [0.97268436, 0.86299324, 0.78630599],
        [0.71481599, 0.98340068, 0.86628930],
        [0.69772882, 0.16384224, 0.17316542],
        [0.21608950, 0.59733394, 0.07494859],
        [0.97627445, 0.00898610, 0.60074272],
        [0.00623026, 0.38657128, 0.16797218],
        [0.25298236, 0.04416006, 0.73338017],
        [0.43479153, 0.95665297, 0.40844386],
    ]
)


class TestRandomTripletGenerator(unittest.TestCase):
    """
    Define :func:`colour.algebra.random.random_triplet_generator` definition
    unit tests methods.
    """

    def test_random_triplet_generator(self):
        """
        Test :func:`colour.algebra.random.random_triplet_generator`
        definition.

        Notes
        -----
        -   The test is assuming that :func:`np.random.RandomState` definition
            will return the same sequence no matter which *OS* or *Python*
            version is used. There is however no formal promise about the
            *prng* sequence reproducibility of either *Python* or *Numpy*
            implementations, see :cite:`Laurent2012a`.
        """

        prng = np.random.RandomState(4)
        np.testing.assert_almost_equal(
            RANDOM_TRIPLETS,
            random_triplet_generator(10, random_state=prng),
            decimal=7,
        )


if __name__ == "__main__":
    unittest.main()



"""Defines the unit tests for the :mod:`colour.utilities.documentation` module."""

import os
import unittest

from colour.utilities.documentation import is_documentation_building

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestIsDocumentationBuilding",
]


class TestIsDocumentationBuilding(unittest.TestCase):
    """
    Define :func:`colour.utilities.documentation.is_documentation_building`
    definition unit tests methods.
    """

    def test_is_documentation_building(self):
        """
        Test :func:`colour.utilities.documentation.is_documentation_building`
        definition.
        """

        try:
            self.assertFalse(is_documentation_building())

            os.environ["READTHEDOCS"] = "True"
            self.assertTrue(is_documentation_building())

            os.environ["READTHEDOCS"] = "False"
            self.assertTrue(is_documentation_building())

            del os.environ["READTHEDOCS"]
            self.assertFalse(is_documentation_building())

            os.environ["COLOUR_SCIENCE__DOCUMENTATION_BUILD"] = "True"
            self.assertTrue(is_documentation_building())

            os.environ["COLOUR_SCIENCE__DOCUMENTATION_BUILD"] = "False"
            self.assertTrue(is_documentation_building())

            del os.environ["COLOUR_SCIENCE__DOCUMENTATION_BUILD"]
            self.assertFalse(is_documentation_building())

        finally:  # pragma: no cover
            if os.environ.get("READTHEDOCS"):
                del os.environ["READTHEDOCS"]

            if os.environ.get("COLOUR_SCIENCE__DOCUMENTATION_BUILD"):
                del os.environ["COLOUR_SCIENCE__DOCUMENTATION_BUILD"]


if __name__ == "__main__":
    unittest.main()



"""
DCI-P3 & DCI-P3+ Colourspaces
=============================

Defines the *DCI-P3* and *DCI-P3+* colourspaces:

-   :attr:`colour.models.RGB_COLOURSPACE_DCI_P3`.
-   :attr:`colour.models.RGB_COLOURSPACE_DCI_P3_P`.

References
----------
-   :cite:`Canon2014a` : Canon. (2014). EOS C500 Firmware Update. Retrieved
    August 27, 2016, from
    https://www.usa.canon.com/internet/portal/us/home/explore/\
product-showcases/cameras-and-lenses/cinema-eos-firmware/c500
-   :cite:`DigitalCinemaInitiatives2007b` : Digital Cinema Initiatives. (2007).
    Digital Cinema System Specification - Version 1.1.
    http://www.dcimovies.com/archives/spec_v1_1/\
DCI_DCinema_System_Spec_v1_1.pdf
-   :cite:`Hewlett-PackardDevelopmentCompany2009a` : Hewlett-Packard
    Development Company. (2009). Understanding the HP DreamColor LP2480zx
    DCI-P3 Emulation Color Space (pp. 1-3).
    http://www.hp.com/united-states/campaigns/workstations/pdfs/\
lp2480zx-dci--p3-emulation.pdf
"""

from __future__ import annotations

import numpy as np
from functools import partial

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import (
    RGB_Colourspace,
    gamma_function,
    normalised_primary_matrix,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_DCI_P3",
    "PRIMARIES_DCI_P3_P",
    "WHITEPOINT_NAME_DCI_P3",
    "CCS_WHITEPOINT_DCI_P3",
    "MATRIX_DCI_P3_TO_XYZ",
    "MATRIX_XYZ_TO_DCI_P3",
    "MATRIX_DCI_P3_P_TO_XYZ",
    "MATRIX_XYZ_TO_DCI_P3_P",
    "RGB_COLOURSPACE_DCI_P3",
    "RGB_COLOURSPACE_DCI_P3_P",
]

PRIMARIES_DCI_P3: NDArray = np.array(
    [
        [0.6800, 0.3200],
        [0.2650, 0.6900],
        [0.1500, 0.0600],
    ]
)
"""*DCI-P3* colourspace primaries."""

PRIMARIES_DCI_P3_P: NDArray = np.array(
    [
        [0.7400, 0.2700],
        [0.2200, 0.7800],
        [0.0900, -0.0900],
    ]
)
"""*DCI-P3+* colourspace primaries."""

WHITEPOINT_NAME_DCI_P3: str = "DCI-P3"
"""
*DCI-P3* colourspace whitepoint name.

Warnings
--------
DCI-P3 illuminant has no associated spectral distribution. DCI has no
official reference spectral measurement for this whitepoint. The closest
matching spectral distribution is Kinoton 75P projector.
"""

CCS_WHITEPOINT_DCI_P3: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_DCI_P3]
"""*DCI-P3* colourspace whitepoint chromaticity coordinates."""

MATRIX_DCI_P3_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_DCI_P3, CCS_WHITEPOINT_DCI_P3
)
"""*DCI-P3* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_DCI_P3: NDArray = np.linalg.inv(MATRIX_DCI_P3_TO_XYZ)
"""*CIE XYZ* tristimulus values to *DCI-P3* colourspace matrix."""

MATRIX_DCI_P3_P_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_DCI_P3_P, CCS_WHITEPOINT_DCI_P3
)
"""*DCI-P3+* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_DCI_P3_P: NDArray = np.linalg.inv(MATRIX_DCI_P3_P_TO_XYZ)
"""*CIE XYZ* tristimulus values to *DCI-P3+* colourspace matrix."""

RGB_COLOURSPACE_DCI_P3: RGB_Colourspace = RGB_Colourspace(
    "DCI-P3",
    PRIMARIES_DCI_P3,
    CCS_WHITEPOINT_DCI_P3,
    WHITEPOINT_NAME_DCI_P3,
    MATRIX_DCI_P3_TO_XYZ,
    MATRIX_XYZ_TO_DCI_P3,
    partial(gamma_function, exponent=1 / 2.6),
    partial(gamma_function, exponent=2.6),
)
RGB_COLOURSPACE_DCI_P3.__doc__ = """
*DCI-P3* colourspace.

References
----------
:cite:`DigitalCinemaInitiatives2007b`,
:cite:`Hewlett-PackardDevelopmentCompany2009a`
"""

RGB_COLOURSPACE_DCI_P3_P: RGB_Colourspace = RGB_Colourspace(
    "DCI-P3+",
    PRIMARIES_DCI_P3_P,
    CCS_WHITEPOINT_DCI_P3,
    WHITEPOINT_NAME_DCI_P3,
    MATRIX_DCI_P3_P_TO_XYZ,
    MATRIX_XYZ_TO_DCI_P3_P,
    partial(gamma_function, exponent=1 / 2.6),
    partial(gamma_function, exponent=2.6),
)
RGB_COLOURSPACE_DCI_P3_P.__doc__ = """
*DCI-P3+* colourspace.

References
----------
:cite:`Canon2014a`
"""



"""Showcases *Luminance* computations."""

import colour

from colour.utilities import message_box

message_box('"Luminance" Computations')

V = 4.08244375
message_box(
    f'Computing "luminance" using "Newhall, Nickerson, and Judd (1943)" '
    f'method for given "Munsell" value:\n\n\t{V}'
)
print(colour.luminance(V, method="Newhall 1943"))
print(colour.colorimetry.luminance_Newhall1943(V))

print("\n")

L = 41.527875844653451
message_box(
    f'Computing "luminance" using "CIE 1976" method for given '
    f'"Lightness":\n\n\t{L}'
)
print(colour.luminance(L))
print(colour.colorimetry.luminance_CIE1976(L))

print("\n")

L = 31.996390226262736
message_box(
    f'Computing "luminance" using "Fairchild and Wyble (2010)" method for '
    f'given "Lightness":\n\n\t{L}'
)
print(colour.luminance(L, method="Fairchild 2010") * 100)
print(colour.colorimetry.luminance_Fairchild2010(L) * 100)

print("\n")

L = 51.852958445912506
message_box(
    f'Computing "luminance" using "Fairchild and Chen (2011)" method for '
    f'given "Lightness":\n\n\t{L}'
)
print(colour.luminance(L, method="Fairchild 2011") * 100)
print(colour.colorimetry.luminance_Fairchild2011(L) * 100)

print("\n")

message_box(
    f'Computing "luminance" using "ASTM D1535-08e1" method for given '
    f'"Munsell" value:\n\n\t{V}'
)
print(colour.luminance(V, method="ASTM D1535"))
print(colour.colorimetry.luminance_ASTMD1535(V))



"""Defines the unit tests for the :mod:`colour.colorimetry.luminance` module."""

import numpy as np
import unittest

from colour.colorimetry import (
    luminance_Newhall1943,
    intermediate_luminance_function_CIE1976,
    luminance_CIE1976,
    luminance_ASTMD1535,
    luminance_Fairchild2010,
    luminance_Fairchild2011,
    luminance_Abebe2017,
)
from colour.colorimetry.luminance import luminance
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestLuminanceNewhall1943",
    "TestLuminanceASTMD1535",
    "TestIntermediateLuminanceFunctionCIE1976",
    "TestLuminanceCIE1976",
    "TestLuminanceFairchild2010",
    "TestLuminanceFairchild2011",
    "TestLuminanceAbebe2017",
    "TestLuminance",
]


class TestLuminanceNewhall1943(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.luminance.luminance_Newhall1943`
    definition unit tests methods.
    """

    def test_luminance_Newhall1943(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Newhall1943`
        definition.
        """

        self.assertAlmostEqual(
            luminance_Newhall1943(4.08244375), 12.550078816731881, places=7
        )

        self.assertAlmostEqual(
            luminance_Newhall1943(5.39132685), 23.481252371310738, places=7
        )

        self.assertAlmostEqual(
            luminance_Newhall1943(2.97619312), 6.4514266875601924, places=7
        )

    def test_n_dimensional_luminance_Newhall1943(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Newhall1943`
        definition n-dimensional arrays support.
        """

        V = 4.08244375
        Y = luminance_Newhall1943(V)

        V = np.tile(V, 6)
        Y = np.tile(Y, 6)
        np.testing.assert_almost_equal(luminance_Newhall1943(V), Y, decimal=7)

        V = np.reshape(V, (2, 3))
        Y = np.reshape(Y, (2, 3))
        np.testing.assert_almost_equal(luminance_Newhall1943(V), Y, decimal=7)

        V = np.reshape(V, (2, 3, 1))
        Y = np.reshape(Y, (2, 3, 1))
        np.testing.assert_almost_equal(luminance_Newhall1943(V), Y, decimal=7)

    def test_domain_range_scale_luminance_Newhall1943(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Newhall1943`
        definition domain and range scale support.
        """

        Y = luminance_Newhall1943(4.08244375)

        d_r = (("reference", 1, 1), ("1", 0.1, 0.01), ("100", 10, 1))
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    luminance_Newhall1943(4.08244375 * factor_a),
                    Y * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_luminance_Newhall1943(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Newhall1943`
        definition nan support.
        """

        luminance_Newhall1943(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLuminanceASTMD1535(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.luminance.luminance_ASTMD1535`
    definition unit tests methods.
    """

    def test_luminance_ASTMD1535(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_ASTMD1535`
        definition.
        """

        self.assertAlmostEqual(
            luminance_ASTMD1535(4.08244375), 12.236342675366036, places=7
        )

        self.assertAlmostEqual(
            luminance_ASTMD1535(5.39132685), 22.893999867280378, places=7
        )

        self.assertAlmostEqual(
            luminance_ASTMD1535(2.97619312), 6.2902253509053132, places=7
        )

    def test_n_dimensional_luminance_ASTMD1535(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_ASTMD1535`
        definition n-dimensional arrays support.
        """

        V = 4.08244375
        Y = luminance_ASTMD1535(V)

        V = np.tile(V, 6)
        Y = np.tile(Y, 6)
        np.testing.assert_almost_equal(luminance_ASTMD1535(V), Y, decimal=7)

        V = np.reshape(V, (2, 3))
        Y = np.reshape(Y, (2, 3))
        np.testing.assert_almost_equal(luminance_ASTMD1535(V), Y, decimal=7)

        V = np.reshape(V, (2, 3, 1))
        Y = np.reshape(Y, (2, 3, 1))
        np.testing.assert_almost_equal(luminance_ASTMD1535(V), Y, decimal=7)

    def test_domain_range_scale_luminance_ASTMD1535(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_ASTMD1535`
        definition domain and range scale support.
        """

        Y = luminance_ASTMD1535(4.08244375)

        d_r = (("reference", 1, 1), ("1", 0.1, 0.01), ("100", 10, 1))
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    luminance_ASTMD1535(4.08244375 * factor_a),
                    Y * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_luminance_ASTMD1535(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_ASTMD1535`
        definition nan support.
        """

        luminance_ASTMD1535(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestIntermediateLuminanceFunctionCIE1976(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.luminance.\
intermediate_luminance_function_CIE1976` definition unit tests methods.
    """

    def test_intermediate_luminance_function_CIE1976(self):
        """
        Test :func:`colour.colorimetry.luminance.\
intermediate_luminance_function_CIE1976` definition.
        """

        self.assertAlmostEqual(
            intermediate_luminance_function_CIE1976(0.495929964178047),
            12.197225350000002,
            places=7,
        )

        self.assertAlmostEqual(
            intermediate_luminance_function_CIE1976(0.613072093530391),
            23.042767810000004,
            places=7,
        )

        self.assertAlmostEqual(
            intermediate_luminance_function_CIE1976(0.394876333449113),
            6.157200790000001,
            places=7,
        )

    def test_n_dimensional_intermediate_luminance_function_CIE1976(self):
        """
        Test :func:`colour.colorimetry.luminance.\
intermediate_luminance_function_CIE1976` definition n-dimensional arrays
        support.
        """

        f_Y_Y_n = 0.495929964178047
        Y = intermediate_luminance_function_CIE1976(f_Y_Y_n)

        f_Y_Y_n = np.tile(f_Y_Y_n, 6)
        Y = np.tile(Y, 6)
        np.testing.assert_almost_equal(
            intermediate_luminance_function_CIE1976(f_Y_Y_n), Y, decimal=7
        )

        f_Y_Y_n = np.reshape(f_Y_Y_n, (2, 3))
        Y = np.reshape(Y, (2, 3))
        np.testing.assert_almost_equal(
            intermediate_luminance_function_CIE1976(f_Y_Y_n), Y, decimal=7
        )

        f_Y_Y_n = np.reshape(f_Y_Y_n, (2, 3, 1))
        Y = np.reshape(Y, (2, 3, 1))
        np.testing.assert_almost_equal(
            intermediate_luminance_function_CIE1976(f_Y_Y_n), Y, decimal=7
        )

    def test_domain_range_scale_intermediate_luminance_function_CIE1976(self):
        """
        Test :func:`colour.colorimetry.luminance.\
intermediate_luminance_function_CIE1976` definition domain and range scale
        support.
        """

        Y = intermediate_luminance_function_CIE1976(41.527875844653451, 100)

        for scale in ("reference", "1", "100"):
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    intermediate_luminance_function_CIE1976(
                        41.527875844653451, 100
                    ),
                    Y,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_intermediate_luminance_function_CIE1976(self):
        """
        Test :func:`colour.colorimetry.luminance.\
intermediate_luminance_function_CIE1976` definition nan support.
        """

        intermediate_luminance_function_CIE1976(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLuminanceCIE1976(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.luminance.luminance_CIE1976` definition
    unit tests methods.
    """

    def test_luminance_CIE1976(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_CIE1976`
        definition.
        """

        self.assertAlmostEqual(
            luminance_CIE1976(41.527875844653451), 12.197225350000002, places=7
        )

        self.assertAlmostEqual(
            luminance_CIE1976(55.116362849525402), 23.042767810000004, places=7
        )

        self.assertAlmostEqual(
            luminance_CIE1976(29.805654680097106), 6.157200790000001, places=7
        )

        self.assertAlmostEqual(
            luminance_CIE1976(56.480581732417676, 50),
            12.197225349999998,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_CIE1976(47.317620274162735, 75),
            12.197225350000002,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_CIE1976(42.519930728120940, 95),
            12.197225350000005,
            places=7,
        )

    def test_n_dimensional_luminance_CIE1976(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_CIE1976`
        definition n-dimensional arrays support.
        """

        L_star = 41.527875844653451
        Y = luminance_CIE1976(L_star)

        L_star = np.tile(L_star, 6)
        Y = np.tile(Y, 6)
        np.testing.assert_almost_equal(luminance_CIE1976(L_star), Y, decimal=7)

        L_star = np.reshape(L_star, (2, 3))
        Y = np.reshape(Y, (2, 3))
        np.testing.assert_almost_equal(luminance_CIE1976(L_star), Y, decimal=7)

        L_star = np.reshape(L_star, (2, 3, 1))
        Y = np.reshape(Y, (2, 3, 1))
        np.testing.assert_almost_equal(luminance_CIE1976(L_star), Y, decimal=7)

    def test_domain_range_scale_luminance_CIE1976(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_CIE1976`
        definition domain and range scale support.
        """

        Y = luminance_CIE1976(41.527875844653451, 100)

        d_r = (("reference", 1), ("1", 0.01), ("100", 1))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    luminance_CIE1976(41.527875844653451 * factor, 100),
                    Y * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_luminance_CIE1976(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_CIE1976`
        definition nan support.
        """

        luminance_CIE1976(np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]))


class TestLuminanceFairchild2010(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.luminance.luminance_Fairchild2010`
    definition unit tests methods.
    """

    def test_luminance_Fairchild2010(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Fairchild2010`
        definition.
        """

        self.assertAlmostEqual(
            luminance_Fairchild2010(31.996390226262736),
            0.12197225350000002,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Fairchild2010(60.203153682783302),
            0.23042767809999998,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Fairchild2010(11.836517240976489),
            0.06157200790000001,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Fairchild2010(24.424283249379986, 2.75),
            0.12197225350000002,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Fairchild2010(100.019986327374240),
            1008.00000024,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Fairchild2010(100.019999997090270),
            100799.92312466,
            places=7,
        )

    def test_n_dimensional_luminance_Fairchild2010(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Fairchild2010`
        definition n-dimensional arrays support.
        """

        L_hdr = 31.996390226262736
        Y = luminance_Fairchild2010(L_hdr)

        L_hdr = np.tile(L_hdr, 6)
        Y = np.tile(Y, 6)
        np.testing.assert_almost_equal(
            luminance_Fairchild2010(L_hdr), Y, decimal=7
        )

        L_hdr = np.reshape(L_hdr, (2, 3))
        Y = np.reshape(Y, (2, 3))
        np.testing.assert_almost_equal(
            luminance_Fairchild2010(L_hdr), Y, decimal=7
        )

        L_hdr = np.reshape(L_hdr, (2, 3, 1))
        Y = np.reshape(Y, (2, 3, 1))
        np.testing.assert_almost_equal(
            luminance_Fairchild2010(L_hdr), Y, decimal=7
        )

    def test_domain_range_scale_luminance_Fairchild2010(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Fairchild2010`
        definition domain and range scale support.
        """

        Y = luminance_Fairchild2010(31.996390226262736)

        d_r = (("reference", 1, 1), ("1", 0.01, 1), ("100", 1, 100))
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    luminance_Fairchild2010(31.996390226262736 * factor_a),
                    Y * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_luminance_Fairchild2010(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Fairchild2010`
        definition nan support.
        """

        luminance_Fairchild2010(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLuminanceFairchild2011(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.luminance.luminance_Fairchild2011`
    definition unit tests methods.
    """

    def test_luminance_Fairchild2011(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Fairchild2011`
        definition.
        """

        self.assertAlmostEqual(
            luminance_Fairchild2011(51.852958445912506),
            0.12197225350000007,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Fairchild2011(65.275207956353853),
            0.23042767809999998,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Fairchild2011(39.818935510715917),
            0.061572007900000038,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Fairchild2011(0.13268968410139345, 2.75),
            0.12197225350000002,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Fairchild2011(234.72925681957565),
            1008.00000000,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Fairchild2011(245.57059778237573),
            100800.00000000,
            places=7,
        )

    def test_n_dimensional_luminance_Fairchild2011(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Fairchild2011`
        definition n-dimensional arrays support.
        """

        L_hdr = 51.852958445912506
        Y = luminance_Fairchild2011(L_hdr)

        L_hdr = np.tile(L_hdr, 6)
        Y = np.tile(Y, 6)
        np.testing.assert_almost_equal(
            luminance_Fairchild2011(L_hdr), Y, decimal=7
        )

        L_hdr = np.reshape(L_hdr, (2, 3))
        Y = np.reshape(Y, (2, 3))
        np.testing.assert_almost_equal(
            luminance_Fairchild2011(L_hdr), Y, decimal=7
        )

        L_hdr = np.reshape(L_hdr, (2, 3, 1))
        Y = np.reshape(Y, (2, 3, 1))
        np.testing.assert_almost_equal(
            luminance_Fairchild2011(L_hdr), Y, decimal=7
        )

    def test_domain_range_scale_luminance_Fairchild2011(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Fairchild2011`
        definition domain and range scale support.
        """

        Y = luminance_Fairchild2011(26.459509817572265)

        d_r = (("reference", 1, 1), ("1", 0.01, 1), ("100", 1, 100))
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    luminance_Fairchild2011(26.459509817572265 * factor_a),
                    Y * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_luminance_Fairchild2011(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Fairchild2011`
        definition nan support.
        """

        luminance_Fairchild2011(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLuminanceAbebe2017(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.luminance.luminance_Abebe2017`
    definition unit tests methods.
    """

    def test_luminance_Abebe2017(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Abebe2017`
        definition.
        """

        self.assertAlmostEqual(
            luminance_Abebe2017(0.486955571109229),
            12.197225350000004,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Abebe2017(0.474544792145434, method="Stevens"),
            12.197225350000025,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Abebe2017(0.286847428534793, 1000),
            12.197225350000046,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Abebe2017(0.192145492588158, 4000),
            12.197225350000121,
            places=7,
        )

        self.assertAlmostEqual(
            luminance_Abebe2017(0.170365211220992, 4000, method="Stevens"),
            12.197225349999933,
            places=7,
        )

    def test_n_dimensional_luminance_Abebe2017(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Abebe2017`
        definition n-dimensional arrays support.
        """

        L = 0.486955571109229
        Y = luminance_Abebe2017(L)

        L = np.tile(L, 6)
        Y = np.tile(Y, 6)
        np.testing.assert_almost_equal(luminance_Abebe2017(L), Y, decimal=7)

        L = np.reshape(L, (2, 3))
        Y = np.reshape(Y, (2, 3))
        np.testing.assert_almost_equal(luminance_Abebe2017(L), Y, decimal=7)

        L = np.reshape(L, (2, 3, 1))
        Y = np.reshape(Y, (2, 3, 1))
        np.testing.assert_almost_equal(luminance_Abebe2017(L), Y, decimal=7)

    def test_domain_range_scale_luminance_Abebe2017(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Abebe2017`
        definition domain and range scale support.
        """

        L = luminance_Abebe2017(0.486955571109229)

        d_r = (("reference", 1), ("1", 1), ("100", 1))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    luminance_Abebe2017(
                        0.486955571109229 * factor, 100 * factor
                    ),
                    L * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_luminance_Abebe2017(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance_Abebe2017`
        definition nan support.
        """

        luminance_Abebe2017(
            *[np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])] * 2
        )


class TestLuminance(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.luminance.luminance` definition unit
    tests methods.
    """

    def test_domain_range_scale_luminance(self):
        """
        Test :func:`colour.colorimetry.luminance.luminance` definition
        domain and range scale support.
        """

        m = (
            "Newhall 1943",
            "ASTM D1535",
            "CIE 1976",
            "Fairchild 2010",
            "Fairchild 2011",
            "Abebe 2017",
        )
        v = [luminance(41.527875844653451, method, Y_n=100) for method in m]

        d_r = (("reference", 1), ("1", 0.01), ("100", 1))
        for method, value in zip(m, v):
            for scale, factor in d_r:
                with domain_range_scale(scale):
                    np.testing.assert_almost_equal(
                        luminance(
                            41.527875844653451 * factor, method, Y_n=100
                        ),
                        value * factor,
                        decimal=7,
                    )


if __name__ == "__main__":
    unittest.main()



"""
Showcases reflectance recovery computations using *Mallett et al. (2019)*
method.
"""

import numpy as np

import colour
from colour.utilities import message_box

message_box('"Mallett et al. (2019)" - Reflectance Recovery Computations')

illuminant = colour.SDS_ILLUMINANTS["D65"]

XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
RGB = colour.XYZ_to_sRGB(XYZ, apply_cctf_encoding=False)
message_box(
    f'Recovering reflectance using "Mallett et al. (2019)" method from given '
    f'"XYZ" tristimulus values:\n\n\tXYZ: {XYZ}'
)
sd = colour.XYZ_to_sd(XYZ, method="Mallett 2019")
print(sd)
print(colour.recovery.RGB_to_sd_Mallett2019(RGB))
print(colour.sd_to_XYZ(sd, illuminant=illuminant) / 100)

print("\n")

message_box(
    'Generating the "Mallett et al. (2019)" basis functions for the '
    "*Pal/Secam* colourspace:"
)
cmfs = (
    colour.MSDS_CMFS["CIE 1931 2 Degree Standard Observer"]
    .copy()
    .align(colour.SpectralShape(360, 780, 10))
)
illuminant = colour.SDS_ILLUMINANTS["D65"].copy().align(cmfs.shape)

print(
    colour.recovery.spectral_primary_decomposition_Mallett2019(
        colour.models.RGB_COLOURSPACE_PAL_SECAM,
        cmfs,
        illuminant,
        optimisation_kwargs={"options": {"ftol": 1e-5}},
    )
)
